// naumann@stce.rwth-aachen.de

#include "Eigen/Dense"
#include "dco.hpp"

#include <iostream>
#include <cmath>
using namespace std;

template<typename T>
using VT=Eigen::VectorX<T>;

template<typename T>
T f(VT<T> x) {
  T v=0;
  int i=0;
  while (i<x.size()) {
    v=v+pow(x(i),2);
    i=i+1;
  }
  return exp(sin(v));
}

template<typename T>
VT<T> df_dx_cfd(VT<T> x) {
  VT<T> dfdx(x.size());
  T delta_x=1e-7;
  int i=0;
  while (i<x.size()) {
    VT<T> xl=x, xr=x;
    xl(i)=xl(i)-delta_x;
    xr(i)=xr(i)+delta_x;
    dfdx(i)=(f(xr)-f(xl))/(2*delta_x);
    i=i+1;
  }
  return dfdx;
}

template<typename T>
VT<T> df_dx_tad(VT<T> x) {
  using TT=dco::gt1s<T>::type;
  int n=x.size();
  VT<TT> x_t(n);
  TT y_t;
  VT<T> dfdx(n);
  int i=0; 
  while (i<n) {
    dco::value(x_t(i))=x(i);
    i=i+1;
  }
  i=0; 
  while (i<n) {
    dco::derivative(x_t(i))=1;
    y_t=f(x_t);
    dfdx(i)=dco::derivative(y_t);
    dco::derivative(x_t(i))=0;
    i=i+1;
  }
  return dfdx;
}

template<typename T>
VT<T> df_dx_aad(VT<T> x) {
  using AT=dco::ga1s<T>::type;
  int n=x.size();
  VT<AT> x_a(n);
  AT y_a;
  VT<T> dfdx(n);
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  int i=0; 
  while (i<n) {
    dco::value(x_a(i))=x(i);
    tape->register_variable(x_a(i));
    i=i+1;
  }
  y_a=f(x_a);
  dco::derivative(y_a)=1;
  tape->interpret_adjoint();
  i=0; 
  while (i<n) {
    dfdx(i)=dco::derivative(x_a(i));
    i=i+1;
  }
  return dfdx;
}

int main() {
  int n=100000000;
  VT<double> x=VT<double>::Random(n);
  // cout << "x=" << x << endl;
  cout << "f=" << f(x) << endl;
  VT<double> dfdx(n);
  // dfdx=df_dx_cfd(x);
  // cout << "dfdx (cfd)=" << dfdx << endl;
  // dfdx=df_dx_tad(x);
  // cout << "dfdx (tad)=" << dfdx << endl;
  dfdx=df_dx_aad(x);
  cout << "dfdx (aad)=" << dfdx(0) << endl;
  return 0;
}


